import numpy as np
import scipy
import pandas as pd
import itertools
import os
from sklearn.preprocessing import minmax_scale
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans
from sklearn import metrics
from .data_loading import PTIterator, CNIterator, Paired_data_iterator, val_PT_construction, val_CN_construction, val_Gene_construction
from .model import GeneSGAN

import rpy2
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri, Formula
from rpy2.robjects.conversion import localconverter
import rpy2.robjects.packages as rpackages
from rpy2.robjects.vectors import StrVector
utils = rpackages.importr('utils')
packages = ('nnet')
nnet = importr('nnet')

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

def find_highest_ari_model(clustering_results):
	"""
	Find one of save models which have the highest average overlap (evaluated by ARI) with all other saved models
	The returned model will be used as a template such that all other models will \
	be reordered to achieve the highest overlap with it.
	:param clustering_results: list, list of clustering results given by all saved models; 
							   length of list equals the number of saved models.
	:return: int, the index of selected model with highest average overlap
	"""
	highest_ari=0
	best_model=0
	for i in range(len(clustering_results)):
		all_ari=[]
		for j in range(len(clustering_results)):
			if i!=j:all_ari.append(metrics.adjusted_rand_score(clustering_results[i], clustering_results[j]))
		if np.mean(all_ari)>highest_ari:
			best_model=i
			highest_ari=np.mean(all_ari)
	return best_model


def get_model_order(cluster_results,ncluster):
	"""
	Find best orders for results given by all saved models so that they reach highest agreements
	:param clustering_results: list, list of clustering results given by all saved models; 
							   length of list equals the number of saved models.
	:param ncluster: int, number of clusters
	:return: list, list of best orders for results given by all saved models.
	"""
	order_permutation = list(itertools.permutations(range(ncluster)))
	best_model = find_highest_ari_model(cluster_results)
	all_orders=[]
	for k in range(len(cluster_results)):
		if k==best_model: 
			all_orders.append(range(ncluster))
		elif k!=best_model:
			highest_intersection = 0
			for order in order_permutation:
				total_intersection = 0
				for i in range(ncluster):
					total_intersection+= np.intersect1d(np.where(cluster_results[best_model]==i),np.where(cluster_results[k]==order[i])).shape[0]
				if total_intersection>=highest_intersection:
					best_order=order
					highest_intersection=total_intersection
			all_orders.append(best_order)
	return all_orders

def highest_matching_clustering(clustering_results, label_probability, ncluster):
	"""
	The function which offers clustering result (cluster label and cluster probabilities) 
	by reordering and combining clustering results given by all saved models
	:param clustering_results: list, list of clustering results given by all saved models; 
							   length of list equals the number of saved models.
	:param label_probability: list, list of clutering results given by all saved modesl;
								length of list equals the number of saved models and each 
								model gives an n*k list.
	:param ncluster: int, number of clusters
	:return: two arrays, one n*1 array with cluster label for each participant
						 one n*k array with k cluster probabilites for each participant
	"""
	order = get_model_order(clustering_results, ncluster)
	class_index=0
	for i in range(len(clustering_results)):
		label_probability[i] = label_probability[i][:,order[i]]
	prediction_prob=np.mean(label_probability,axis=0)
	prediction_cluster=prediction_prob.argmax(axis=1)
	return prediction_cluster, prediction_prob, 

def Covariate_correction(cn_data,cn_cov,pt_data,pt_cov):
	"""
	Eliminate the confound of covariate, such as age and sex, from the disease-based changes.
	:param cn_data: array, control data
	:param cn_cov: array, control covariates
	:param pt_data: array, patient data
	:param pt_cov: array, patient covariates
	:return: corrected control data & corrected patient data
	"""
	min_cov = np.amin(cn_cov, axis=0)
	max_cov = np.amax(cn_cov, axis=0)
	pt_cov = (pt_cov-min_cov)/(max_cov-min_cov)
	cn_cov = (cn_cov-min_cov)/(max_cov-min_cov)
	beta = np.transpose(LinearRegression().fit(cn_cov, cn_data).coef_)
	corrected_cn_data = cn_data-np.dot(cn_cov,beta)
	corrected_pt_data = pt_data-np.dot(pt_cov,beta)
	correction_variables = {'max_cov':max_cov,'min_cov':min_cov,'beta':beta}
	return corrected_cn_data, corrected_pt_data, correction_variables

def Data_normalization(cn_data,pt_data):
	"""
	Normalize all data with respect to control data to ensure a mean of 1 and std of 0.1 
	among CN participants for each ROI
	:param cn_data: array, control data
	:param pt_data: array, patient data
	:return: normalized control data & normalized patient data
	"""
	cn_mean = np.mean(cn_data,axis=0)
	cn_std = np.std(cn_data,axis=0)
	normalized_cn_data = 1+(cn_data-cn_mean)/(10*cn_std)
	normalized_pt_data = 1+(pt_data-cn_mean)/(10*cn_std)
	normalization_variables = {'cn_mean':cn_mean, 'cn_std':cn_std}
	return normalized_cn_data, normalized_pt_data, normalization_variables

def apply_covariate_correction(data, covariate, correction_variables):
	covariate = (covariate-correction_variables['min_cov'])/(correction_variables['max_cov']-correction_variables['min_cov'])
	corrected_data = data-np.dot(covariate,correction_variables['beta'])
	return corrected_data

def apply_data_normalization(data, normalization_variables):
	normalized_data = 1+(data-normalization_variables['cn_mean'])/(10*normalization_variables['cn_std'])
	return normalized_data

def parse_train_data(imaging_data, gene_data, covariate, random_seed, data_fraction, batch_size):
	np.random.seed(random_seed)
	n_pt_data = imaging_data.loc[imaging_data['diagnosis'] != -1].shape[0]
	n_cn_data = imaging_data.loc[imaging_data['diagnosis'] == -1].shape[0]
	indices_pt = np.random.choice(n_pt_data, int(data_fraction*n_pt_data), replace=False)
	np.random.seed(random_seed)
	indices_cn = np.random.choice(n_cn_data, int(data_fraction*n_cn_data), replace=False)
	cn_data = imaging_data.loc[imaging_data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values[indices_cn]
	pt_data = imaging_data.loc[imaging_data['diagnosis'] !=- 1].drop(['participant_id','diagnosis'], axis=1).values[indices_pt]
	gene_data = imaging_data[['participant_id']].merge(gene_data,on='participant_id')
	gene_data_fillna = gene_data.fillna(gene_data.mean())
	gene_data = gene_data.drop(['participant_id'], axis=1).values[indices_pt]
	gene_data_fillna = gene_data_fillna.drop(['participant_id'], axis=1).values[indices_pt]
	correction_variables = None
	if covariate is not None:
		cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values[indices_cn]
		pt_cov = covariate.loc[covariate['diagnosis'] != -1].drop(['participant_id','diagnosis'], axis=1).values[indices_pt]
		cn_data,pt_data,correction_variables = Covariate_correction(cn_data,cn_cov,pt_data,pt_cov)
	normalized_cn_data, normalized_pt_data,normalization_variables = Data_normalization(cn_data,pt_data)
	cn_train_dataset = CNIterator(normalized_cn_data, batch_size)
	pt_train_dataset = PTIterator(normalized_pt_data, batch_size)
	pair_train_dataset = Paired_data_iterator(gene_data, gene_data_fillna, normalized_pt_data,  data_fraction, batch_size, random_seed)
	return cn_train_dataset, pt_train_dataset, pair_train_dataset, correction_variables, normalization_variables

def parse_validation_data(imaging_data, gene_data, covariate, random_seed, data_fraction, correction_variables, normalization_variables):
	cn_data = imaging_data.loc[imaging_data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values
	pt_data = imaging_data.loc[imaging_data['diagnosis'] != -1].drop(['participant_id','diagnosis'], axis=1).values
	gene_data = gene_data.drop(['participant_id'], axis=1).values
	if covariate is not None:
		cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values
		pt_cov = covariate.loc[covariate['diagnosis'] != -1].drop(['participant_id','diagnosis'], axis=1).values
		cn_data = apply_covariate_correction(cn_data, cn_cov, correction_variables)
		pt_data = apply_covariate_correction(pt_data, pt_cov, correction_variables)
	normalized_cn_data = apply_data_normalization(cn_data, normalization_variables)
	normalized_pt_data = apply_data_normalization(pt_data, normalization_variables)
	cn_train_dataset = val_CN_construction(normalized_cn_data, random_seed, data_fraction).load_train()
	pt_train_dataset = val_PT_construction(normalized_pt_data, random_seed, data_fraction).load_train()
	gene_train_dataset = val_Gene_construction(gene_data, random_seed, data_fraction,).load_train()
	pt_test_dataset = val_PT_construction(normalized_pt_data, random_seed, data_fraction).load_test()
	gene_test_dataset = val_Gene_construction(gene_data, random_seed, data_fraction,).load_test()
	return cn_train_dataset, pt_train_dataset, gene_train_dataset, pt_test_dataset, gene_test_dataset

def mnlogit(test_dataframe):
	with localconverter(ro.default_converter+pandas2ri.converter):
		r_dataframe_test = ro.conversion.py2rpy(test_dataframe)
	if test_dataframe.columns.shape[0]>1:
		formula = 'label~'+test_dataframe.columns[0]
		for i in range(1,test_dataframe.columns.shape[0]-1):
			formula+="+"+test_dataframe.columns[i]
	else:
		formula = 'label~1'
	fmla = Formula(formula)
	model = nnet.multinom(fmla,data=r_dataframe_test,trace=False)
	pred_prob = np.array(ro.r.predict(model,type="probs",newdata = r_dataframe_test))
	return -metrics.log_loss(test_dataframe['label'],pred_prob,normalize=False)


def n_sig_calculation(model_dir, image_data, covariate_data, gene_data, fraction):
	n_sig_snp = []
	fold_number = len(model_dir)
	if covariate_data is not None:
		pt_cov = covariate_data.loc[covariate_data['diagnosis'] != -1].drop(['participant_id','diagnosis'], axis=1).values
	else:
		pt_cov = None
	n_pt = gene_data.values.shape[0]
	i=0
	for saved_model in model_dir:
		i+=1
		np.random.seed(i)
		indices = np.random.choice(n_pt, int(fraction*n_pt), replace=False)
		select = np.in1d(range(n_pt), indices)
		if pt_cov is not None:
			test_cov = pt_cov[~select]
		else:
			test_cov = None
		model = GeneSGAN()
		model.load(saved_model)
		cn_data, train_pt_data, train_gene_data, val_pt_data, val_gene_data = parse_validation_data(image_data, gene_data, covariate_data, i, fraction, model.opt.correction_variables, model.opt.normalization_variables)
		test_label = np.argmax(model.predict_cluster(val_pt_data), axis=1)
		n_sig_snp.append(likelihood_ratio_n(test_label,val_gene_data,test_cov))
	return np.mean(n_sig_snp), np.std(n_sig_snp)


def likelihood_ratio_n(test_label, test_snp, test_cov):
	if test_cov is not None:
		test_cov = test_cov.astype(float)
		test_cov  = (test_cov -np.mean(test_cov ,axis=0))/np.std(test_cov ,axis=0)
		test_model_data = pd.DataFrame(data=test_cov,columns=['covariate'+str(_) for _ in range(test_cov.shape[1])])
		test_model_data['label']=test_label
	else:
		test_model_data = pd.DataFrame()
		test_model_data['label']=test_label
	reduce_llf = mnlogit(test_model_data)
	all_ll_difference = []
	n = 0
	for i in range(test_snp.shape[1]):
		if test_cov is not None:
			X_full_test = np.concatenate((test_cov,test_snp[:,i].numpy().reshape(-1,1)),axis=1)
			test_model_data = pd.DataFrame(data=X_full_test,columns=['covariate'+str(_) for _ in range(test_cov.shape[1])]+['snp'])
		else:
			test_model_data = pd.DataFrame(data=test_snp[:,i].numpy().reshape(-1,1),columns=['snp'])
		test_model_data['label']=test_label
		test_model_data = test_model_data.dropna()
		full_llf = mnlogit(test_model_data)
		all_ll_difference.append(full_llf-reduce_llf)
		if full_llf-reduce_llf>=3.84:
			n+=1
	return n






import numpy as np
from scipy.linalg import sqrtm
import torch
from .model import sample_z_categorical, sample_z_continuous

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

## calulate wasserstein distance assuming independent ROIs
def cal_w_distance(mean_1,mean_2,std_1,std_2):
    distance=0
    for i in range(mean_1.size):
        distance+=((mean_1[i]-mean_2[i])**2+(std_1[i])**2+std_2[i]**2-2*std_1[i]*std_2[i])
    return  distance


## return mean wasserstein distance and wasserstein distance of each mapping direction
def cal_validate_distance(model,mean,std,eva_data):
    predicted_z=model.predict_cluster(eva_data)
    cluster=[[]for k in range(model.opt.ncluster)]
    distance=[float('inf') for m in range(model.opt.ncluster)]
    adjusted_eva_data = eva_data.detach().numpy()
    quantity = [0 for _ in range(model.opt.ncluster)]
    for i in range(predicted_z.shape[0]):
        a = predicted_z[i].tolist().index(max(predicted_z[i].tolist()))
        cluster[a].append(adjusted_eva_data[i])
        quantity[a]+=1
    for j in range(model.opt.ncluster):
        if len(cluster[j])>1:
            cluster_mean=np.mean(np.array(cluster[j]), axis=0)
            cluster_std=np.std(np.array(cluster[j]), axis=0)
            distance[j]=cal_w_distance(mean[j],cluster_mean,std[j],cluster_std)
    output=[float('inf') for _ in range(model.opt.ncluster)]
    total_distance = 0
    for k in range(model.opt.ncluster):
        total_distance += quantity[k]*distance[k]
        if distance[k]!=float('inf'):
            output[k] = distance[k]
    if np.max(quantity)/np.sum(quantity)>0.9 or np.min(quantity)==0:
        return float('inf'), output
    else:
        return total_distance/np.sum(quantity), output
      
def eval_w_distances(real_X,real_Y,model):
    predict_Y=[]
    mean=[]
    std=[]
    for i in range(model.opt.ncluster):
        z1, z1_index = sample_z_categorical(real_X, model.opt.ncluster, fix_class=i)
        z2 = sample_z_continuous(real_X, model.opt.n_z2)
        z1_2=torch.cat((z1,z2),1)
        predict_Y.append(model.predict_Y(real_X,z1_2).detach().numpy())
    for j in range(model.opt.ncluster):
        mean.append(np.mean(predict_Y[j], axis=0))
        std.append(np.std(predict_Y[j], axis=0))
    mean_distance, w_distances=cal_validate_distance(model,mean,std,real_Y)
    return mean_distance, w_distances

## return clustering memberships of valY datapoints and quantity of subjects assigned to each subtype
def label_change(model,test_Y,opt):
    predicted_label=[]
    predicted_class=[0 for _ in range(opt.ncluster)]
    prediction=model.predict_cluster(test_Y)
    for i in range(prediction.shape[0]):
        label=prediction[i].tolist().index(max(prediction[i].tolist()))
        predicted_label.append(label)
        predicted_class[label]+=1
    return predicted_label, predicted_class









import sys
import os
import numpy as np
from sklearn import metrics
from .model import GeneSGAN
from .utils import highest_matching_clustering, parse_validation_data, n_sig_calculation
from .clustering import Gene_SGAN_train

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"


def model_filtering(model_dirs, ncluster, imaging_data, gene_data, covariate=None):
	"""
	Function used for filter out models who have significantly different clustering results with others.
	This function deal with rare failing cases of Gene-SGAN
	Args:
		model_dirs: list, list of dirs of all saved models
		ncluster: int, number of defined clusters
		data, data_frame, dataframe with same format as training data. CN data must be exactly same as CN data in training dataframe while 
						  PT data can be any samples in or out of the training set.
		covariate, data_frame, dataframe with same format as training covariate. CN data must be exactly same as CN data in training covariate while 
						  PT data can be any samples in or out of the training set.

	Returns: list of index indicating outlier models

	"""
	
	all_prediction_labels = []
	for models in model_dirs:
		model = GeneSGAN()
		model.load(models)
		__, validation_data, __, __, __ = parse_validation_data(imaging_data, gene_data, covariate, 1, 1, model.opt.correction_variables, model.opt.normalization_variables)
		all_prediction_labels.append(np.argmax(model.predict_cluster(validation_data), axis=1))
	model_aris = [[] for _ in range(len(model_dirs))]
	filtered_models = []

	for i in range(len(model_dirs)):
		for j in range(len(model_dirs)):
			if i!=j:
				model_aris[i].append(metrics.adjusted_rand_score(all_prediction_labels[i], all_prediction_labels[j]))
	median_aris = np.median(model_aris, axis=1)
	for j in range(median_aris.shape[0]):
		rest_aris = np.delete(median_aris,j)
		if (median_aris[j]-np.mean(rest_aris))/np.std(rest_aris)<-2:
			filtered_models.append(j)

	return filtered_models

def calculate_ari(prediction_labels):
	model_aris = []
	for i in range(len(prediction_labels)):
		for j in range(i+1,len(prediction_labels)):
			model_aris.append(metrics.adjusted_rand_score(prediction_labels[i], prediction_labels[j]))
	return np.mean(model_aris), np.std(model_aris)



def clustering_result(model_dirs, ncluster, imaging_data, gene_data, fraction=0.8, covariate=None):
	"""
	Function used for derive clustering results from several saved models
	Args:
		model_dirs: list, list of dirs of all saved models
		ncluster: int, number of defined clusters
		consensus_type: string, the method used for deriving final clustering results with all models derived through CV
								choose between 'highest_matching_clustering' and 'consensus_clustering'
		data, data_frame, dataframe with same format as training data. CN data must be exactly same as CN data in training dataframe while 
						  PT data can be any samples in or out of the training set.
		covariate, data_frame, dataframe with same format as training covariate. CN data must be exactly same as CN data in training covariate while 
						  PT data can be any samples in or out of the training set.
	Returns: clustering outputs.

	"""

	all_prediction_labels = []
	all_prediction_probabilities = []
	for models in model_dirs:
		model = GeneSGAN()
		model.load(models)
		__, validation_data, __, __, __ = parse_validation_data(imaging_data, gene_data, covariate, 1, 1, model.opt.correction_variables, model.opt.normalization_variables)
		all_prediction_labels.append(np.argmax(model.predict_cluster(validation_data), axis=1))
		all_prediction_probabilities.append(model.predict_cluster(validation_data))

	mean_ari, std_ari = calculate_ari(all_prediction_labels)
	print("Results have Adjuested_random_index (ARI) = %.2f+- %.2f" %(mean_ari, std_ari))

	mean_n_sig_snps, std_n_sig_snps = n_sig_calculation(model_dirs,imaging_data,covariate,gene_data, fraction)
			
	cluster_label, cluster_prob = highest_matching_clustering(all_prediction_labels, all_prediction_probabilities, ncluster)
	return cluster_label, cluster_prob, mean_ari, std_ari, mean_n_sig_snps, std_n_sig_snps


def cross_validated_clustering(imaging_data, gene_data, ncluster, fold_number, fraction, start_saving_epoch, max_epoch, output_dir, WD_threshold, AQ_threshold, \
		cluster_loss_threshold, covariate=None, lam=9, mu=5, kappa=0.1, batchSize=25, lipschitz_k = 0.5, verbose = False, \
		beta1 = 0.5, lr = 0.001, genelr = 0.0002, max_gnorm = 100, eval_freq = 5, save_epoch_freq = 5, start_fold = 0, stop_fold = None, check_outlier = True):
	"""
	cross_validated clustering function using Gene-SGAN (recommended)
	Args:
		imaging_data: dataframe, dataframe file with all imaging (phenotypic) features. The dataframe contains
		the following headers: "
								 "i) the first column is the participant_id;"
								 "ii) the second column should be the diagnosis;"
								 "The following column should be the extracted phenotypic features."
		gene_data: dataframe, dataframe file with all genetic features of patients. The dataframe contains
		the following headers: "
								 "i) the first column is the participant_id of patients;"
								 "The following column should be the extracted genetic features."
		covariate: dataframe, not required; dataframe file with all confounding covariates to be corrected. The dataframe contains
		the following headers: "
								 "i) the first column is the participant_id;"
								 "ii) the second column should be the diagnosis;"
								 "The following column should be all confounding covariates. e.g., age, sex"
		ncluster: int, number of defined clusters
		fold_number: int, number of folds for leave-out cross validation
		fraction: float, fraction of data used for training in each fold
		start_saving_epoch: int, epoch number from which model will be saved and training will be stopped if stopping criteria satisfied
		max_epoch: int, maximum trainig epoch: training will stop even if criteria not satisfied.
		output_dir: str, the directory underwhich model and results will be saved
		WD_threshold: int, chosen WD theshold for stopping criteria
		AQ_threshold: int, chosen AQ threhold for stopping criteria
		cluster_loss_threshold: int, chosen cluster_loss threhold for stopping criteria
		saved_model_name: str, the name of the saved model
		lam: int, hyperparameter for cluster loss
		mu: int, hyperparameter for change loss
		batchsize: int, batck size for training procedure
		lipschitz_k = float, hyper parameter for weight clipping of mapping and clustering function
		verbose: bool, choose whether to print out training procedure
		beta1: float, parameter of ADAM optimization method
		lr: float, learning rate of phenotype step
		genelr: float, learning rate of gene step
		max_gnorm: float, maximum gradient norm for gradient clipping
		eval_freq: int, the frequency at which the model is evaluated during training procedure
		save_epoch_freq: int, the frequency at which the model is saved during training procedure
		start_fold; int, indicate the last saved fold index,
							  used for restart previous half-finished cross validation; set defaultly to be 0 indicating a new cv process
		stop_fold: int, indicate the index of fold at which the cv early stop,
							  used for stopping cv process eartly and resuming later; set defaultly to be None and cv will not stop till the end
		check_outlier: bool, whether check outlier model (potential unsuccessful model) after cv process and retrain the fold

	Returns: clustering outputs.

	"""
	
	print('Start Gene-SGAN for semi-supervised clustering')

	Gene_SGAN_model = Gene_SGAN_train(ncluster, start_saving_epoch, max_epoch, WD_threshold, AQ_threshold, \
		cluster_loss_threshold, lam=lam, mu=mu, kappa = kappa, batchSize=batchSize, \
		lipschitz_k = lipschitz_k, beta1 = beta1, lr = lr, genelr = genelr, max_gnorm = max_gnorm, eval_freq = eval_freq, save_epoch_freq = save_epoch_freq)

	saved_models = [os.path.join(output_dir, 'converged_model_fold'+str(i)) for i in range(fold_number)]
	
	if stop_fold == None:
		stop_fold = fold_number
	for i in range(start_fold, stop_fold):
		print('****** Starting training of Fold '+str(i)+" ******")
		saved_model_name = 'converged_model_fold'+str(i)
		converge = Gene_SGAN_model.train(saved_model_name, imaging_data, gene_data, covariate, output_dir, random_seed=i, data_fraction = fraction, verbose = verbose)
		while not converge:
			print("****** Model not converging or not converged at max interation, Start retraining ******")
			converge = Gene_SGAN_model.train(saved_model_name, imaging_data, gene_data, covariate, output_dir, random_seed=i, data_fraction = fraction, verbose = verbose)

	if check_outlier:
		print('****** Start Checking outlier models ******')
		outlier_models = model_filtering(saved_models, ncluster, imaging_data, gene_data, covariate)
		if len(outlier_models) > 0:
			print('Model', end=' ')
			for model in outlier_models:
				print(str(model),end=' ')
			print('have low agreement with other models')
		else:
			print('****** There are no outlier models ******')
	
		for i in outlier_models:
			print('****** Starting training of Fold '+str(i)+" ******")
			saved_model_name = 'converged_model_fold'+str(i)
			converge = Gene_SGAN_model.train(saved_model_name, imaging_data, gene_data, covariate, output_dir, verbose = verbose)
			while not converge:
				print("****** Model not converged at max interation, Start retraining ******")
				converge = Gene_SGAN_model.train(saved_model_name, imaging_data, gene_data, covariate, output_dir, random_seed=i, data_fraction = fraction, verbose = verbose)

	cluster_label, cluster_prob, mean_ari, std_ari, mean_n_sig_snp, std_n_sig_snp = clustering_result(saved_models, ncluster, imaging_data, gene_data, fraction = fraction, covariate = covariate)
	
	pt_data = imaging_data.loc[imaging_data['diagnosis'] != -1][['participant_id','diagnosis']]
	pt_data['cluster_label'] = cluster_label + 1

	for i in range(ncluster):
		pt_data['p'+str(i+1)] = cluster_prob[:,i]

	pt_data["ARI = %.2f+- %.2f" %(mean_ari, std_ari)] = ''
	pt_data["N-Asso-SNPs = %.2f+- %.2f" %(mean_n_sig_snp, std_n_sig_snp)] = ''
	
	pt_data.to_csv(os.path.join(output_dir,'clustering_result.csv'), index = False)
	print('****** Gene-SGAN clustering finished ******')



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as D
from .modules import TwoInputSequential, Sub_Adder

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

###############################################################################
# Functions
###############################################################################

def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
       m.weight.data.normal_(0, 0.12)

def define_Linear_Transform(nROI,nCluster,nLatent):
    netG = LTransformGenerator(nCluster,nLatent, nROI)
    netG.apply(weights_init)
    return netG

def define_Linear_Discriminator(nROI):
    netD=LDiscriminator(nROI)
    netD.apply(weights_init)
    return netD

def define_Linear_Inverse(nROI,nCluster,nLatent):
    netC = LInverse(nCluster,nLatent, nROI)
    netC.apply(weights_init)
    return netC

def define_Gene_Inference(nGene,nLatent,nCluster):
    netG = gene_inference(nGene,nLatent,nCluster)
    netG.apply(weights_init)
    return netG

def define_z3_Posterior(nGene,nROI,nLatent):
    netI = z3_posterir(nGene,nROI,nLatent)
    netI.apply(weights_init)
    return netI

##############################################################################
# Network Classes
##############################################################################

class LTransformGenerator(nn.Module):
    def __init__(self, nCluster, nLatent, nROI, product_layer=Sub_Adder):
        super(LTransformGenerator, self).__init__()
        model=[]
        def block(in_layer, out_layer, normalize=False):
            layers = [nn.Linear(in_layer, out_layer,bias=False)]
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            return layers
        model+=block(nROI,int(nROI/2))+block(int(nROI/2), int(nROI/4))
        model.append(product_layer(int(nROI/4),nCluster+nLatent))
        model+=block(int(nROI/4),int(nROI/2))+block(int(nROI/2),nROI)
        model.append(nn.Linear(nROI, nROI,bias=False))
        self.model = TwoInputSequential(*model)
    def forward(self, input_x,input_z):
        return self.model(input_x,input_z)

class LInverse(nn.Module):
    def __init__(self, nCluster, nLatent, nROI):
        super(LInverse, self).__init__()
        model=[]
        self.nLatent = nLatent
        def block(in_layer, out_layer, normalize=True):
            layers=[nn.LeakyReLU(0.2, inplace=True)]
            layers.append(nn.Linear(in_layer, out_layer))
            return layers
        model.append(nn.Linear(nROI, nROI))
        model+=block(nROI, int(nROI/2), normalize=False)\
            +block(int(nROI/2), int(nROI/4))\
            +block(int(nROI/4), nCluster+nLatent)
        self.model = nn.Sequential(*model)

    def forward(self, input_y):
        z = self.model(input_y)
        z2 = z[:, 0:self.nLatent]
        z1 = z[:, self.nLatent:]
        z1_softmax = F.softmax(z1, dim=1)
        return z1_softmax, z1, z2

class LDiscriminator(nn.Module):
    def __init__(self, nROI):
        super(LDiscriminator, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(nROI, int(nROI/2),bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(int(nROI/2), int(nROI/4),bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(int(nROI/4), 2,bias=True),
        )
    def forward(self, input_y):
        pred = self.model(input_y)
        return pred

class z3_posterir(nn.Module):
    def __init__(self,nGene,nROI,nLatent):
        super(z3_posterir, self).__init__()
        self.model=nn.Sequential(nn.Linear((nGene+nROI),int((nGene+nROI)/2)),
                                    nn.LeakyReLU(0.2, inplace=True),
                                    nn.Linear(int((nGene+nROI)/2),int((nGene+nROI)/4)),
                                    nn.LeakyReLU(0.2, inplace=True),
                                    nn.Linear(int((nGene+nROI)/4), nLatent*2))

    def forward(self, input_x):
        mu, logsigma = self.model(input_x).chunk(2, dim=-1)
        return D.Normal(mu, abs(logsigma.exp()))

class gene_inference(nn.Module):
    def __init__(self,nGene,nLatent,nCluster):
        super(gene_inference, self).__init__()
        self.model=nn.Sequential(nn.Linear((nCluster+nLatent),2*(nCluster+nLatent)),
                                    nn.LeakyReLU(0.2, inplace=True),
                                    nn.Linear(2*(nCluster+nLatent),4*(nCluster+nLatent)),
                                    nn.LeakyReLU(0.2, inplace=True),
                                    nn.Dropout(p=0.2),
                                    nn.Linear(4*(nCluster+nLatent), nGene))

    def forward(self, input_y,input_z):
        yz = torch.cat([input_y, input_z], dim=1)
        logits = self.model(yz)
        return D.Binomial(total_count=2,logits=logits), torch.sigmoid(logits)





import os
import torch
import torch.nn.functional as F
import torch.distributions as D
from collections import OrderedDict
from torch.autograd import Variable
from itertools import chain as ichain
from .networks import define_Linear_Transform, define_Linear_Inverse, define_Linear_Discriminator, define_Gene_Inference, define_z3_Posterior

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"


#####sample from discrete uniform random variable and construct SUB variable. 
def sample_z_categorical(x, ncluster, fix_class=-1):
	"""
	sample from discrete uniform random variable and construct SUB variable.

	:param x, torch tensor with real data 
	:param ncluster: int, defined number of clusters
	:param fix_class: int, can be set to certain mapping directions to generate data in only one cluster,
						   set to -1 for random sampling from discrete uniform distribution

	:return tensor with sampled or selected mapping directions & 
			tensors with shape n*k (each row is a one-hot vector depending on sampled directions) 
	"""
	Tensor = torch.FloatTensor
	z1 = Tensor(x.size(0), ncluster).fill_(0)
	z1_idx = torch.empty(x.size(0), dtype=torch.long)
	if (fix_class == -1):
		z1_idx = z1_idx.random_(ncluster)
		z1 = z1.scatter_(1, z1_idx.unsqueeze(1), 1.)
	else:
		z1_idx[:] = fix_class
		z1 = z1.scatter_(1, z1_idx.unsqueeze(1), 1.)
		z1[0,fix_class]
	z1_var = Variable(z1)
	return z1_var, z1_idx

def sample_z_continuous(x, n_z2):
	z2=Variable(x.data.new(x.size(0),n_z2).uniform_(0,1))
	return z2

def criterion_GAN(pred, target_is_real, prob):
	if target_is_real:
		target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(0.))
		loss=(F.cross_entropy(pred, target_var, reduce=False).unsqueeze(-1)*prob).mean()
	else:
		target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(1.))
		loss = (F.cross_entropy(pred, target_var, reduce=False).unsqueeze(-1)*prob).mean()
	return loss

def criterion_VAE(v, z, p_z, p_v_z, q_z_v):
		v_nan_mask = ~v.isnan()
		v = torch.nan_to_num(v,nan=0)
		return (- torch.masked_select(p_v_z.log_prob(v),v_nan_mask).sum() \
			   + q_z_v.log_prob(z).sum()\
			   - p_z.log_prob(z).sum())/v.shape[0]

def KL_divergence(P,Q):
	return (P*(torch.log(P)-torch.log(Q))).sum()

class dotdict(dict):
	"""dot.notation access to dictionary attributes"""
	__getattr__ = dict.get
	__setattr__ = dict.__setitem__
	__delattr__ = dict.__delitem__

class GeneSGAN(object):
	def __init__(self):
		self.opt = None

		##### definition of all netwotks
		self.netMapping = None
		self.netClustering= None
		self.netDiscriminator = None

		##### definition of all optimizers
		self.optimizer_M = None
		self.optimizer_D = None

		##### definition of all criterions
		self.criterionGAN = criterion_GAN
		self.criterionChange = F.l1_loss
		self.criterionCluster = F.cross_entropy
		self.criterionLatent = F.mse_loss
		self.criterionVAE = criterion_VAE


	def create(self, opt):

		self.opt = opt

		## definition of networkds for GAN
		self.netTransform = define_Linear_Transform(self.opt.nROI,self.opt.ncluster,self.opt.n_z2)
		self.netInverse = define_Linear_Inverse(self.opt.nROI,self.opt.ncluster,self.opt.n_z2)
		self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI)
		self.phi = torch.nn.Parameter(torch.ones(self.opt.ncluster)/self.opt.ncluster)

		### definition of networkds for VAE
		self.netGeneInference = define_Gene_Inference(self.opt.nGene,self.opt.n_z3,self.opt.ncluster)
		self.netGeneZ3Posterior = define_z3_Posterior(self.opt.nGene,self.opt.nROI,self.opt.n_z3)

		## definition of all optimizers
		self.optimizer_M = torch.optim.Adam(ichain(self.netTransform.parameters(),self.netInverse.parameters()),
										lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
		self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
											lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))
		self.optimizer_Gene_VAE = torch.optim.Adam(ichain(self.netGeneInference.parameters(),self.netInverse.parameters(),self.netGeneZ3Posterior.parameters()),
											lr=self.opt.genelr)
		self.optimizer_phi = torch.optim.Adam(iter([self.phi]),
											lr=self.opt.lr/40., betas=(self.opt.beta1, 0.999))

		##### definition of prior probability
		self.p_z3 = D.Normal(torch.tensor(0.), torch.tensor(1.))


	def train_instance(self, x, real_y, pair_gene, pair_gene_filled, pair_y, kappa):

		initial = torch.ones(self.opt.ncluster)/self.opt.ncluster
		z1, z1_index = sample_z_categorical(x, self.opt.ncluster)
		z2 = sample_z_continuous(x, self.opt.n_z2)
		z_1_2 = torch.cat((z1,z2),1)
		fake_y = self.netTransform.forward(x,z_1_2)+x
		softmax_phi = F.softmax(self.phi, dim=0)
		post_prob = torch.gather(torch.stack([softmax_phi for _ in range(z1_index.shape[0])], dim=0),1,z1_index.unsqueeze(-1))

		## Discriminator loss
		pred_fake_y = self.netDiscriminator.forward(fake_y.detach())
		loss_D_fake_y = self.criterionGAN(pred_fake_y, False,post_prob.detach())
		pred_true_y = self.netDiscriminator.forward(real_y)
		loss_D_true_y = self.criterionGAN(pred_true_y, True,post_prob.detach())
		loss_D= 0.5* (loss_D_fake_y + loss_D_true_y)

		## update weights of discriminator
		self.optimizer_D.zero_grad()
		loss_D.backward()
		gnorm_D = torch.nn.utils.clip_grad_norm_(self.netDiscriminator.parameters(), self.opt.max_gnorm)
		self.optimizer_D.step()

		## Phi loss
		loss_phi = self.criterionGAN(self.netDiscriminator.forward(fake_y.detach()).detach(), True, post_prob)+kappa*KL_divergence(initial, softmax_phi)

		self.optimizer_phi.zero_grad()
		loss_phi.backward()
		self.optimizer_phi.step()

		## GAN/Change/Recons loss
		pred_fake_y = self.netDiscriminator.forward(fake_y)
		loss_GAN = self.criterionGAN(pred_fake_y, True, post_prob.detach())

		reconst_z1_softmax, reconst_z1, reconst_z2 = self.netInverse.forward(fake_y)
		z1_loss = self.criterionCluster(reconst_z1, z1_index)
		z2_loss = self.criterionLatent(reconst_z2, z2)

		change_loss= self.criterionChange(fake_y, x)
		loss_G = (1/torch.mean(post_prob.detach()))*loss_GAN+self.opt.lam*(z1_loss+z2_loss)+self.opt.mu*change_loss

		## update weights of Transformation and Inverse function
		self.optimizer_M.zero_grad()
		loss_G.backward()
		gnorm_M = torch.nn.utils.clip_grad_norm_(self.netTransform.parameters(), self.opt.max_gnorm)
		self.optimizer_M.step()

		##########################
		#VAE step for Genetic Data
		##########################
		estimated_z1, __ , __ = self.netInverse.forward(pair_y)

		pair_v_y = torch.cat([(pair_gene_filled-1.0)/5., pair_y], dim=1)
		q_z3 = self.netGeneZ3Posterior.forward(pair_v_y)
		z3 = q_z3.rsample()
		p_infered_v, v_prob = self.netGeneInference.forward(estimated_z1, z3)
		VAE_loss = self.criterionVAE(pair_gene, z3, self.p_z3, p_infered_v, q_z3)

		self.optimizer_Gene_VAE.zero_grad()
		VAE_loss.backward()
		self.optimizer_Gene_VAE.step()

		losses=OrderedDict([('Vae_loss', VAE_loss.item()),('Discriminator_loss', loss_D.item()),('Mapping_loss', loss_GAN.item()),('loss_change', change_loss.item()),('loss_cluster', z1_loss.item())])

		## perform weight clipping
		for p in self.netTransform.parameters():
			p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)
		for p in self.netInverse.parameters():
			p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)

		return losses
	
	## return a numpy array of pattern type probabilities for all input subjects
	def predict_cluster(self,real_y):
		prediction,z1,z2=self.netInverse.forward(real_y)
		return prediction.detach().numpy()

	## return generated patient data with given zub variable
	def predict_Y(self, x, z):
		return self.netTransform.forward(x, z)+x

	def predict_gene_prob(self, z1, z2):
		return self.netGeneInference.forward(z1, z2)[0].probs

	def get_phi(self):
		return F.softmax(self.phi.detach(), dim=0)

	## save checkpoint    
	def save(self, save_dir, chk_name):
		chk_path = os.path.join(save_dir, chk_name)
		checkpoint = {
			'netTransform':self.netTransform.state_dict(),
			'netDiscriminator':self.netDiscriminator.state_dict(),
			'optimizer_D':self.optimizer_D.state_dict(),
			'optimizer_M':self.optimizer_M.state_dict(),
			'netInverse':self.netInverse.state_dict(),
			'netGeneInference':self.netGeneInference.state_dict(),
			'netZLatentGene':self.netGeneZ3Posterior.state_dict(),
			'optimizer_Gene_VAE':self.optimizer_Gene_VAE.state_dict(),
			'optimizer_phi':self.optimizer_phi.state_dict()
		}
		checkpoint.update(self.opt)
		torch.save(checkpoint, chk_path)

	def load_opt(self,checkpoint):
		self.opt = dotdict({})
		for key in checkpoint.keys():
			if key not in ['netTransform','netDiscriminator','netInverse','netGeneInference','netZLatentGene','optimizer_M','optimizer_D','optimizer_Gene_VAE','optimizer_phi']:
				self.opt[key] = checkpoint[key]
		

	## load trained model
	def load(self, chk_path):
		checkpoint = torch.load(chk_path)
		self.load_opt(checkpoint)

		## definition of networkds for GAN
		self.netTransform = define_Linear_Transform(self.opt.nROI,self.opt.ncluster,self.opt.n_z2)
		self.netInverse = define_Linear_Inverse(self.opt.nROI,self.opt.ncluster,self.opt.n_z2)
		self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI)
		self.phi = torch.nn.Parameter(torch.ones(self.opt.ncluster)/self.opt.ncluster)

		### definition of networkds for VAE
		self.netGeneInference = define_Gene_Inference(self.opt.nGene,self.opt.n_z3,self.opt.ncluster)
		self.netGeneZ3Posterior = define_z3_Posterior(self.opt.nGene,self.opt.nROI,self.opt.n_z3)


		## definition of all optimizers
		self.optimizer_M = torch.optim.Adam(ichain(self.netTransform.parameters(),self.netInverse.parameters()),
										lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
		self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
											lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))
		self.optimizer_Gene_VAE = torch.optim.Adam(ichain(self.netGeneInference.parameters(),self.netInverse.parameters(),self.netGeneZ3Posterior.parameters()),
											lr=self.opt.genelr)
		self.optimizer_phi = torch.optim.Adam(iter([self.phi]),
											lr=self.opt.lr/40., betas=(self.opt.beta1, 0.999))

		##### definition of prior probability
		self.p_z3 = D.Normal(torch.tensor(0.), torch.tensor(1.))
		
		self.netTransform.load_state_dict(checkpoint['netTransform'])
		self.netDiscriminator.load_state_dict(checkpoint['netDiscriminator'])
		self.netInverse.load_state_dict(checkpoint['netInverse'])
		self.optimizer_D.load_state_dict(checkpoint['optimizer_D'])
		self.optimizer_M.load_state_dict(checkpoint['optimizer_M'])
		self.netGeneInference.load_state_dict(checkpoint['netGeneInference'])
		self.netGeneZ3Posterior.load_state_dict(checkpoint['netZLatentGene'])
		self.optimizer_Gene_VAE.load_state_dict(checkpoint['optimizer_Gene_VAE'])
		self.optimizer_phi.load_state_dict(checkpoint['optimizer_phi'])
		self.load_opt(checkpoint)
			


		
